//
// Copyright 2014, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.
#ifndef vector_vector_sort_header_included_p
#define vector_vector_sort_header_included_p

#include <parse_date.h>
#include <stdio.h>
#include <string>
#include <vector>

///@file 
/// 
///  This file declares classes in the cxxtls::vectorVector namespace.
///  This namespace exists to define tools used in comparing, and
///  perhaps manipulating, vectors of vectors of other things.
///

namespace cxxtls
{
namespace vectorVector
   ///  A collection of definitions which are used to compare and/or
   ///  manipulate vectors of vectors.
{

#define VECTOR_VECTOR_OUTRAGEOUSLY_SMALL (-1e30)

struct SortKey
   ///  This class represents a comparator for a std::string stored in a vector.
   ///  there are various modes of comparison:  std::string, number, date. 
   ///
   ///  Note that this comparator assumes that leading and trailing blanks
   ///  have been trimmed earlier in the processing.
{
    
    enum sortMode
    {
      SK_string,
      SK_number,
      SK_date
    };

    size_t     index_;
    sortMode   mode_;
    bool       invert_;

    SortKey()
    :  index_(0)
    ,  mode_(SK_string)
    ,  invert_(false)
    {
    }

    SortKey(size_t index, sortMode mode, bool invert=false)
    :  index_(index)
    ,  mode_(mode)
    ,  invert_(invert)
    {
    }

    bool compareLessHelper(std::string const &lhs, std::string const &rhs) const
        ///  @return true if lhs is less than rhs given the current
        ///          sort mode
    {
        if(mode_ == SK_string)
            return lhs < rhs;             // std::string comparison
        else
        if(mode_ == SK_date)
        {
           return dateComparison(lhs,rhs);  // date comparison
        }

        double lh = VECTOR_VECTOR_OUTRAGEOUSLY_SMALL;
        double rh = VECTOR_VECTOR_OUTRAGEOUSLY_SMALL;

        sscanf(lhs.c_str(), "%lf", &lh);
        sscanf(rhs.c_str(), "%lf", &rh);

        // if conversion failure, then lh and rh will remain outrageously small

        return lh < rh;

    }

    bool compareLess(std::string const &lhs, std::string const &rhs) const
    {
        bool less = compareLessHelper(lhs,rhs);

        if(invert_)
           less = !less;

        return less;
    }


    static bool dateComparison(std::string const &lhs, std::string const &rhs)
        /// generalized time and date comparison.  Dates must obey the
        /// rules described in the help page.
    {
        time_t lh = parse_date(lhs.c_str());
        time_t rh = parse_date(rhs.c_str());

        return lh < rh;
        
    }

};

struct Sorter
   /// A wrapper around a composite key that applies to vectors of strings.
   /// That is, a collection of SortKey along with the fields that they apply
   /// to.
{
      std::vector<SortKey> key_;

      size_t minIndex_;

      void push_back(SortKey const &k)
      {
         if(k.index_ > minIndex_)
            minIndex_ = k.index_;

         key_.push_back(k);
      }

      bool operator() (std::vector<std::string> const &lhs, std::vector<std::string> const &rhs) const
      {
          // we are assuming lhs and rhs are big enough for the keys to be
          // accessed.  This should be inforced at the top level algorithm.

          std::vector<SortKey>::const_iterator f = key_.begin(), l = key_.end();

          while(f != l)
          {
              SortKey const &key = *f++;

              bool less =  key.compareLess(lhs[key.index_], rhs[key.index_]);

              if(less)
              {
                 // the keys are processed in order of importance.  So, as soon as you 
                 // find one that indicates less, that means the whole comparison returns less.

                 return true;   
              }


              // if we get here lhs >= rhs

              less =  key.compareLess(rhs[key.index_], lhs[key.index_]);

              if(less)
                 {
                    // as soon as you find one string that is greater, then the whole
                    // comparison returns greater.

                    return false; // lhs is clearly bigger
                 }

              // if we get here lhs == rhs

              // keep going to the next key

          }

          // if we get here, then the keys are equal.

          return false;
      }

      bool empty() const { return key_.empty(); }

      Sorter()
      :  minIndex_(0)
      {
      }

};

} // vectorVector

} // namespace cxxtls

#endif
