#ifndef cxxtls_line_stream_header_included_p
#define cxxtls_line_stream_header_included_p
//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <cxxtls/CommandConsole.h>

#include <string>
#include <vector>
#include <list>
#include <sstream>

#include <utility>
#include <memory>

#ifdef _MSC_VER
#include <windows.h>
#else
#include <unistd.h>
#endif

///@file
///  This file defines a couple of classes related to reading command lines from
///  file handles.  See ScriptInputLine, ScriptLineStream, and the helper
///  class AutoPoppingPushPrompt.


namespace cxxtls
{

    struct ScriptInputLine
       /// this class is a representation of the line of input from a stream.  It contains
       /// a string member, test_, which holds the text of the line.  It also contains
       /// filename_ and lineNumber_ which reflect the source of the xt.
    {
        std::string text_;
        std::string filename_;
        size_t      lineNumber_;

        ScriptInputLine(std::string const &filename, size_t lineNumber)
        :  filename_(filename)
        ,  lineNumber_(lineNumber)
        {
        }

        ScriptInputLine(std::string const &filename, size_t lineNumber, std::string const &text)
        :  text_(text)
        ,  filename_(filename)
        ,  lineNumber_(lineNumber)
        {
        }

        operator std::string const &() const { return text_; } ///<  conversion operator that lets
                                                               ///   you treat a ScriptInputLine
                                                               ///   as if it were just a string.


        ScriptInputLine()
        :  filename_("")
        ,  lineNumber_(0)
        {
        }

    };

    class ScriptLineStream
        /// A wrapper around a file handle that lets you read ScriptInputLine objects from
        /// it.  This includes lines of unlimited lenght.  It also handles the prompting
        /// of the user with a user specified prompt string.  There is a stack of prompt
        /// strings, and you can push and pop the stack.
    {
        std::string filename_;
        bool        good_;          // not error and not end of file
        bool        error_;         // i/o error occured
        bool        isConsole_;     // based on the file handle
        bool        closeAtExit_;  
        size_t      lineNumber_;
        char        buffer_[8192];  // partial lines left over from previous read
        size_t      bytesInBuffer_; // count of bytes in buffer_ left over from previous read
        std::string workingDirectory_; // current directory for tab command interpretation when using consoles

        std::auto_ptr<CommandConsole> console_;

#ifdef _MSC_VER
        typedef HANDLE handle_t;
#else
        typedef int    handle_t;
#endif

        handle_t       fh_;


        std::list<std::string> promptStack_;  // tail is the current prompt

        std::list<std::string> lineGroup_;    // The current group of user input lines, bundled together.
                                              // strings should end in eol

        std::vector<std::string> history_;      // the console reader will call newGroup() to copy lineGroup_
                                                // to old group, and clear lineGroup_.
                                                // This lets the fix command get to the old group.
                                                //  strings should end in eol


        bool recordHistory_;                  // defaults false

        std::stringstream *stringFile_;       // used but not owned
        bool stringEofEncountered_;           // stringFile_ and we have ran out of data
                                              

        void setDefaults()
           // called by constructors to set the nominal state
        {
           filename_             = "stdin";
           lineNumber_           = 0;
           isConsole_            = false;
           bytesInBuffer_        = 0;
           good_                 = false;
           closeAtExit_          = false;
           error_                = true;
           recordHistory_        = false;
           stringEofEncountered_ = false;
        }

        void openCommandConsole(); ///<  connect to console that knows about the screen cursor position stuff

        bool readLineStringStream(ScriptInputLine &scriptInputLine);
            // helper used by readline when dealing with string streams


    public:

        ScriptLineStream(std::string const &filename, std::string const &prompt);  ///<  open specified file and set prompt
        ScriptLineStream(handle_t, bool closeAtExit, std::string const &prompt);   ///<  set up to read the handle and set the prompt
        ScriptLineStream(std::string const &prompt);                               ///<  open stdin as the stream 
        ScriptLineStream(std::stringstream &input);                                ///<  read from a string

        ~ScriptLineStream();

        void setWorkingDirectory(std::string const &curdir)
        {
            workingDirectory_ = curdir;
        }

        void pushPromptStack(std::string const &newPrompt)
        {
           promptStack_.push_back(newPrompt);
        }

        void changePrompt(std::string const &newPrompt)
        {
           if(promptStack_.empty())
              promptStack_.push_back(newPrompt);
           else
              promptStack_.back() = newPrompt;
        }

        void popPromptStack()
        {
            if(!promptStack_.empty())
               promptStack_.pop_back();
        }

        bool eof() const 
        { 
            return !good_; 
        }

        bool error() const
        {
           return error_;
        }

        std::string getPrompt() const 
        { 
            return promptStack_.empty() ? std::string() : promptStack_.back(); 
        }

        bool readLine(ScriptInputLine &scriptInputLine);
            ///<  Read a line from the file or console.  Note:  the input line
            ///   probably has a \n in it.  The only reason that it would not
            ///   is if the input file doesn't end in newline.
            ///   @return true if the data is available.  
            ///   @return false on end of file or error.
            ///
            ///   If recordLineGroups_ is true, then append each red line
            ///   to the lineGroup_ list.

        std::string const &filename()   const { return filename_; }
        int                lineNumber() const { return lineNumber_; }

        bool readFile(ScriptInputLine &scriptInputLine);
            ///<  Read the rest of the file into a single string and
            ///   store it in the scrptInputLine.text_ field as normal.
            ///   
            ///   @return true if the data is available.  
            ///   @return false on end of file or error.


        void setFilename(std::string const &n) { filename_ = n; }  

        bool isConsole() const { return isConsole_; }  ///< return true if the input stream is the console


        void recordHistory(bool value=true)
           // assert that readLine should append its results to the lineGroup_ variable
        {
           recordHistory_ = value;
        }

        void saveHistory();
           ///< Save the current group of lines and make it availble to the lineGroup() function.
           ///  Also, clear the current line group.

        void addToHistory(std::string const &s)
           /// add a newline terminated line to the history
        {
           if(s.empty())
              return;   // don't push empty liens into the history

           history_.push_back(s);
        }

        std::vector< std::string > const & history() const 
           /// get the saved line group (not the current line)
        {
           return history_;
        }

        void clearGroup()
        {
           lineGroup_.clear();
        }

        void addToGroup(std::string const &line)
        {
           if(line.empty())
              return;   // no point in adding blank lines

           lineGroup_.push_back(line);
        }

    };

    struct AutoPoppingPushPrompt
    ///  Upon construction, it pushes the constructor parameter on the stream's prompt stack.
    ///  Upon destruction, it pops the stack.
    {
        ScriptLineStream &stream_;

        AutoPoppingPushPrompt(ScriptLineStream &stream, std::string const &prompt)
        :  stream_(stream)
        {
            stream_.pushPromptStack(prompt);
        }

        ~ AutoPoppingPushPrompt()
        {
            stream_.popPromptStack();
        }
    };


} // cxxtls

#endif
